diff --git a/stratosphere/fatal/source/fatal_config.cpp b/stratosphere/fatal/source/fatal_config.cpp
index 11cc32420..9cc7d91fc 100644
--- a/stratosphere/fatal/source/fatal_config.cpp
+++ b/stratosphere/fatal/source/fatal_config.cpp
@@ -58,9 +58,9 @@ static void SetupConfigLanguages() {
config->error_desc = u8"Please call 1-800-875-1852 for service.\n";
} else {
config->error_desc = u8"An error has occured.\n\n"
- u8"Please press the POWER Button to restart the console, or a VOL button\n"
- u8"to restart the console in RCM mode. If you are unable to restart the\n"
- u8"console, hold the POWER Button for 12 seconds to turn the console off.\n\n"
+ u8"Please press the POWER Button to restart the console normally, or a VOL button\n"
+ u8"to reboot to a payload (or RCM, if none is present). If you are unable to\n"
+ u8"restart the console, hold the POWER Button for 12 seconds to turn the console off.\n\n"
u8"If the problem persists, refer to the Nintendo Support Website.\n"
u8"support.nintendo.com/switch/error\n";
}
diff --git a/stratosphere/fatal/source/fatal_main.cpp b/stratosphere/fatal/source/fatal_main.cpp
index 733dc6eff..39e5ae088 100644
--- a/stratosphere/fatal/source/fatal_main.cpp
+++ b/stratosphere/fatal/source/fatal_main.cpp
@@ -29,6 +29,7 @@
#include "fatal_config.hpp"
#include "fatal_repair.hpp"
#include "fatal_font.hpp"
+#include "fatal_payload_manager.hpp"
extern "C" {
extern u32 __start__;
@@ -159,6 +160,9 @@ int main(int argc, char **argv)
/* Load settings from set:sys. */
InitializeFatalConfig();
+ /* Load a payload from the SD card. */
+ FatalPayloadManager::LoadPayloadFromSdCard();
+
/* Load shared font. */
if (R_FAILED(FontManager::InitializeSharedFont())) {
std::abort();
diff --git a/stratosphere/fatal/source/fatal_payload_manager.cpp b/stratosphere/fatal/source/fatal_payload_manager.cpp
new file mode 100644
index 000000000..52a72d7d6
--- /dev/null
+++ b/stratosphere/fatal/source/fatal_payload_manager.cpp
@@ -0,0 +1,65 @@
+/*
+ * Copyright (c) 2018 Atmosphère-NX
+ *
+ * This program is free software; you can redistribute it and/or modify it
+ * under the terms and conditions of the GNU General Public License,
+ * version 2, as published by the Free Software Foundation.
+ *
+ * This program is distributed in the hope it will be useful, but WITHOUT
+ * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+ * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
+ * more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program. If not, see .
+ */
+
+#include
+#include
+#include
+#include "fatal_types.hpp"
+#include "fatal_payload_manager.hpp"
+
+/* TODO: Find a way to pre-populate this with the contents of fusee-primary. */
+static u8 g_reboot_payload[IRAM_PAYLOAD_MAX_SIZE] __attribute__ ((aligned (0x1000)));
+static u8 g_work_page[0x1000] __attribute__ ((aligned (0x1000)));
+static bool g_payload_loaded = false;
+
+void FatalPayloadManager::LoadPayloadFromSdCard() {
+ FILE *f = fopen("sdmc:/atmosphere/reboot_payload.bin", "rb");
+ if (f == NULL) {
+ return;
+ }
+ ON_SCOPE_EXIT { fclose(f); };
+
+ memset(g_reboot_payload, 0xFF, sizeof(g_reboot_payload));
+ fread(g_reboot_payload, 1, IRAM_PAYLOAD_MAX_SIZE, f);
+ g_payload_loaded = true;
+}
+
+static void ClearIram() {
+ /* Make page FFs. */
+ memset(g_work_page, 0xFF, sizeof(g_work_page));
+
+ /* Overwrite all of IRAM with FFs. */
+ for (size_t ofs = 0; ofs < IRAM_PAYLOAD_MAX_SIZE; ofs += sizeof(g_work_page)) {
+ CopyToIram(IRAM_PAYLOAD_BASE + ofs, g_work_page, sizeof(g_work_page));
+ }
+}
+
+void FatalPayloadManager::RebootToPayload() {
+ /* If we don't actually have a payload loaded, just go to RCM. */
+ if (!g_payload_loaded) {
+ RebootToRcm();
+ }
+
+ /* Ensure clean IRAM state. */
+ ClearIram();
+
+ /* Copy in payload. */
+ for (size_t ofs = 0; ofs < sizeof(g_reboot_payload); ofs += 0x1000) {
+ CopyToIram(IRAM_PAYLOAD_BASE + ofs, &g_reboot_payload[ofs], 0x1000);
+ }
+
+ RebootToIramPayload();
+}
\ No newline at end of file
diff --git a/stratosphere/fatal/source/fatal_payload_manager.hpp b/stratosphere/fatal/source/fatal_payload_manager.hpp
new file mode 100644
index 000000000..cb052b37f
--- /dev/null
+++ b/stratosphere/fatal/source/fatal_payload_manager.hpp
@@ -0,0 +1,28 @@
+/*
+ * Copyright (c) 2018 Atmosphère-NX
+ *
+ * This program is free software; you can redistribute it and/or modify it
+ * under the terms and conditions of the GNU General Public License,
+ * version 2, as published by the Free Software Foundation.
+ *
+ * This program is distributed in the hope it will be useful, but WITHOUT
+ * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+ * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
+ * more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program. If not, see .
+ */
+
+#pragma once
+#include
+#include
+
+#define IRAM_PAYLOAD_MAX_SIZE 0x30000
+#define IRAM_PAYLOAD_BASE 0x40010000ull
+
+class FatalPayloadManager {
+ public:
+ static void LoadPayloadFromSdCard();
+ static void RebootToPayload();
+};
\ No newline at end of file
diff --git a/stratosphere/fatal/source/fatal_task_screen.cpp b/stratosphere/fatal/source/fatal_task_screen.cpp
index 144935406..a67cf71ce 100644
--- a/stratosphere/fatal/source/fatal_task_screen.cpp
+++ b/stratosphere/fatal/source/fatal_task_screen.cpp
@@ -215,9 +215,9 @@ Result ShowFatalTask::ShowFatal() {
} else {
/* Print a special message for atmosphere version mismatch. */
FontManager::Print(u8"Atmosphère version mismatch detected.\n\n"
- u8"Please press the POWER Button to restart the console, or a VOL button\n"
- u8"to restart the console in RCM mode. If you are unable to restart the\n"
- u8"console, hold the POWER Button for 12 seconds to turn the console off.\n\n"
+ u8"Please press the POWER Button to restart the console normally, or a VOL button\n"
+ u8"to reboot to a payload (or RCM, if none is present). If you are unable to\n"
+ u8"restart the console, hold the POWER Button for 12 seconds to turn the console off.\n\n"
u8"Please ensure that all Atmosphère components are updated.\n"
u8"github.com/Atmosphere-NX/Atmosphere/releases\n");
}
diff --git a/stratosphere/libstratosphere b/stratosphere/libstratosphere
index de4c2ddae..9bf228de3 160000
--- a/stratosphere/libstratosphere
+++ b/stratosphere/libstratosphere
@@ -1 +1 @@
-Subproject commit de4c2ddae12207e8ebe5f2f8d40335f0ebd9858d
+Subproject commit 9bf228de3e1e33ea62ed18b51853e5de6b4c4b41
diff --git a/stratosphere/fatal/source/fatal_task_power.cpp b/stratosphere/fatal/source/fatal_task_power.cpp
index 41954a0f5..4883b78f7 100644
--- a/stratosphere/fatal/source/fatal_task_power.cpp
+++ b/stratosphere/fatal/source/fatal_task_power.cpp
@@ -16,6 +16,7 @@
#include
#include "fatal_task_power.hpp"
+#include "fatal_payload_manager.hpp"
#include "fatal_config.hpp"
bool PowerControlTask::TryShutdown() {
@@ -123,13 +124,13 @@ void PowerButtonObserveTask::WaitForPowerButton() {
Result rc = 0;
if (check_vol_up && R_SUCCEEDED((rc = gpioPadGetValue(&vol_up_btn, &val))) && val == GpioValue_Low) {
- /* Tell exosphere to reboot to RCM. */
- RebootToRcm();
+ /* Tell exosphere to reboot to payload. */
+ FatalPayloadManager::RebootToPayload();
}
if (check_vol_down && R_SUCCEEDED((rc = gpioPadGetValue(&vol_down_btn, &val))) && val == GpioValue_Low) {
- /* Tell exosphere to reboot to RCM. */
- RebootToRcm();
+ /* Tell exosphere to reboot to payload. */
+ FatalPayloadManager::RebootToPayload();
}
if ((R_SUCCEEDED(rc = bpcGetSleepButtonState(&state)) && state == BpcSleepButtonState_Held) || (config->quest_flag && reboot_helper.TimedOut())) {